# Copyright (c) 2014 Alex Meade.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# Copyright (c) 2016 Mike Rooney. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from unittest import mock
import uuid

import ddt
from lxml import etree
import six

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap.client import (
    fakes as fake_client)
import cinder.tests.unit.volume.drivers.netapp.dataontap.fakes as fake
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp.dataontap.client import client_base


CONNECTION_INFO = {'hostname': 'hostname',
                   'transport_type': 'https',
                   'port': 443,
                   'username': 'admin',
                   'password': 'passw0rd',
                   'api_trace_pattern': 'fake_regex'}


@ddt.ddt
class NetAppBaseClientTestCase(test.TestCase):

    def setUp(self):
        super(NetAppBaseClientTestCase, self).setUp()

        self.mock_object(client_base, 'LOG')
        self.mock_object(client_base.Client, '_init_ssh_client')
        self.client = client_base.Client(**CONNECTION_INFO)
        self.client.connection = mock.MagicMock()
        self.client.connection.get_api_version.return_value = (1, 100)
        self.client.ssh_client = mock.MagicMock()
        self.connection = self.client.connection
        self.fake_volume = six.text_type(uuid.uuid4())
        self.fake_lun = six.text_type(uuid.uuid4())
        self.fake_size = '1024'
        self.fake_metadata = {'OsType': 'linux', 'SpaceReserved': 'true'}
        self.mock_send_request = self.mock_object(
            self.client.connection, 'send_request')

    def test_get_ontapi_version(self):
        version_response = netapp_api.NaElement(
            etree.XML("""<results status="passed">
                            <major-version>1</major-version>
                            <minor-version>19</minor-version>
                          </results>"""))
        self.connection.invoke_successfully.return_value = version_response

        major, minor = self.client.get_ontapi_version(cached=False)

        self.assertEqual('1', major)
        self.assertEqual('19', minor)

    def test_get_ontapi_version_cached(self):

        self.connection.get_api_version.return_value = (1, 20)

        major, minor = self.client.get_ontapi_version()

        self.assertEqual(1, self.connection.get_api_version.call_count)
        self.assertEqual(1, major)
        self.assertEqual(20, minor)

    def test_check_is_naelement(self):

        element = netapp_api.NaElement('name')

        self.assertIsNone(self.client.check_is_naelement(element))
        self.assertRaises(ValueError, self.client.check_is_naelement, None)

    @ddt.data({'ontap_version': '9.4', 'space_reservation': 'true'},
              {'ontap_version': '9.4', 'space_reservation': 'false'},
              {'ontap_version': '9.6', 'space_reservation': 'true'},
              {'ontap_version': '9.6', 'space_reservation': 'false'})
    @ddt.unpack
    def test_create_lun(self, ontap_version, space_reservation):
        expected_path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        self.fake_metadata['SpaceReserved'] = space_reservation
        expected_space_reservation = space_reservation
        self.mock_object(self.client, 'get_ontap_version',
                         return_value=ontap_version)
        mock_resize_lun = self.mock_object(
            client_base.Client, 'do_direct_resize')
        mock_set_space_reservation = self.mock_object(
            client_base.Client, 'set_lun_space_reservation')
        initial_size = self.fake_size

        if ontap_version < '9.5':
            initial_size = fake.MAX_SIZE_FOR_A_LUN
            expected_space_reservation = 'false'

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.create_lun(self.fake_volume,
                                   self.fake_lun,
                                   self.fake_size,
                                   self.fake_metadata)

            mock_create_node.assert_called_with(
                'lun-create-by-size',
                **{'path': expected_path,
                   'size': initial_size,
                   'ostype': self.fake_metadata['OsType'],
                   'space-reservation-enabled':
                       expected_space_reservation})
            self.connection.invoke_successfully.assert_called_with(
                mock.ANY, True)

        if ontap_version < '9.5':
            mock_resize_lun.assert_called_once_with(
                expected_path, self.fake_size)

        if ontap_version < '9.5' and space_reservation == 'true':
            mock_set_space_reservation.assert_called_once_with(
                expected_path, True)
        else:
            mock_set_space_reservation.assert_not_called()

    @ddt.data({'ontap_version': '9.4', 'space_reservation': 'true'},
              {'ontap_version': '9.4', 'space_reservation': 'false'},
              {'ontap_version': '9.6', 'space_reservation': 'true'},
              {'ontap_version': '9.6', 'space_reservation': 'false'})
    @ddt.unpack
    def test_create_lun_exact_size(self, ontap_version, space_reservation):
        expected_path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        self.connection.get_api_version.return_value = (1, 110)
        self.fake_metadata['SpaceReserved'] = space_reservation
        expected_space_reservation = self.fake_metadata['SpaceReserved']
        self.mock_object(self.client, 'get_ontap_version',
                         return_value=ontap_version)
        mock_resize_lun = self.mock_object(
            client_base.Client, 'do_direct_resize')
        mock_set_space_reservation = self.mock_object(
            client_base.Client, 'set_lun_space_reservation')
        initial_size = self.fake_size

        if ontap_version < '9.5':
            initial_size = fake.MAX_SIZE_FOR_A_LUN
            expected_space_reservation = 'false'

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.create_lun(self.fake_volume,
                                   self.fake_lun,
                                   self.fake_size,
                                   self.fake_metadata)

            mock_create_node.assert_called_with(
                'lun-create-by-size',
                **{'path': expected_path,
                   'size': initial_size,
                   'ostype': self.fake_metadata['OsType'],
                   'use-exact-size': 'true',
                   'space-reservation-enabled':
                       expected_space_reservation})
            self.connection.invoke_successfully.assert_called_with(
                mock.ANY, True)

        if ontap_version < '9.5':
            mock_resize_lun.assert_called_once_with(
                expected_path, self.fake_size)

        if ontap_version < '9.5' and space_reservation == 'true':
            mock_set_space_reservation.assert_called_once_with(
                expected_path, True)
        else:
            mock_set_space_reservation.assert_not_called()

    @ddt.data({'ontap_version': '9.4', 'space_reservation': 'true'},
              {'ontap_version': '9.4', 'space_reservation': 'false'},
              {'ontap_version': '9.6', 'space_reservation': 'true'},
              {'ontap_version': '9.6', 'space_reservation': 'false'})
    @ddt.unpack
    def test_create_lun_with_qos_policy_group_name(
            self, ontap_version, space_reservation):
        expected_path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        expected_qos_group_name = 'qos_1'
        mock_request = mock.Mock()

        self.fake_metadata['SpaceReserved'] = space_reservation
        expected_space_reservation = self.fake_metadata['SpaceReserved']
        self.mock_object(self.client, 'get_ontap_version',
                         return_value=ontap_version)
        mock_resize_lun = self.mock_object(
            client_base.Client, 'do_direct_resize')
        mock_set_space_reservation = self.mock_object(
            client_base.Client, 'set_lun_space_reservation')
        initial_size = self.fake_size

        if ontap_version < '9.5':
            initial_size = fake.MAX_SIZE_FOR_A_LUN
            expected_space_reservation = 'false'

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               return_value=mock_request
                               ) as mock_create_node:
            self.client.create_lun(
                self.fake_volume,
                self.fake_lun,
                self.fake_size,
                self.fake_metadata,
                qos_policy_group_name=expected_qos_group_name)

            mock_create_node.assert_called_with(
                'lun-create-by-size',
                **{'path': expected_path, 'size': initial_size,
                    'ostype': self.fake_metadata['OsType'],
                    'space-reservation-enabled':
                        expected_space_reservation})
            mock_request.add_new_child.assert_called_with(
                'qos-policy-group', expected_qos_group_name)
            self.connection.invoke_successfully.assert_called_with(
                mock.ANY, True)

        if ontap_version < '9.5':
            mock_resize_lun.assert_called_once_with(
                expected_path, self.fake_size)

        if ontap_version < '9.5' and space_reservation == 'true':
            mock_set_space_reservation.assert_called_once_with(
                expected_path, True)
        else:
            mock_set_space_reservation.assert_not_called()

    def test_get_ontap_version(self):
        version_response = netapp_api.NaElement(
            fake.SYSTEM_GET_VERSION_RESPONSE)
        self.connection.invoke_successfully.return_value = version_response

        result = self.client.get_ontap_version(cached=False)

        self.assertEqual(('9.6'), result)

    def test_get_ontap_version_cached(self):
        self.connection.get_ontap_version.return_value = '9.6'

        result = self.client.get_ontap_version()

        self.connection.get_ontap_version.assert_called_once_with()
        self.assertEqual(('9.6'), result)

    def test_set_lun_space_reservation(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_set_space_reservation:
            self.client.set_lun_space_reservation(path, True)

            mock_set_space_reservation.assert_called_once_with(
                'lun-set-space-reservation-info',
                **{'path': path,
                   'enable': 'True'})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    @ddt.data('9.4', '9.6')
    def test_create_lun_raises_on_failure(self, ontap_version):
        self.connection.invoke_successfully = mock.Mock(
            side_effect=netapp_api.NaApiError)
        self.mock_object(self.client, 'get_ontap_version',
                         return_value=ontap_version)

        self.assertRaises(netapp_api.NaApiError,
                          self.client.create_lun,
                          self.fake_volume,
                          self.fake_lun,
                          self.fake_size,
                          self.fake_metadata)

    def test_destroy_lun(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.destroy_lun(path)

            mock_create_node.assert_called_once_with(
                'lun-destroy',
                **{'path': path})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_destroy_lun_force(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        mock_request = mock.Mock()

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               return_value=mock_request
                               ) as mock_create_node:
            self.client.destroy_lun(path)

            mock_create_node.assert_called_once_with('lun-destroy',
                                                     **{'path': path})
            mock_request.add_new_child.assert_called_once_with('force', 'true')
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_map_lun(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        expected_lun_id = 'my_lun'
        mock_response = mock.Mock()
        self.connection.invoke_successfully.return_value = mock_response
        mock_response.get_child_content.return_value = expected_lun_id

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            actual_lun_id = self.client.map_lun(path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-map',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)
            self.assertEqual(expected_lun_id, actual_lun_id)

    def test_map_lun_with_lun_id(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        expected_lun_id = 'my_lun'
        mock_response = mock.Mock()
        self.connection.invoke_successfully.return_value = mock_response
        mock_response.get_child_content.return_value = expected_lun_id

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            actual_lun_id = self.client.map_lun(path, igroup,
                                                lun_id=expected_lun_id)

            mock_create_node.assert_called_once_with(
                'lun-map',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)
            self.assertEqual(expected_lun_id, actual_lun_id)

    def test_map_lun_with_api_error(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        self.connection.invoke_successfully.side_effect =\
            netapp_api.NaApiError()

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.assertRaises(netapp_api.NaApiError, self.client.map_lun,
                              path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-map',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_unmap_lun(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        mock_response = mock.Mock()
        self.connection.invoke_successfully.return_value = mock_response

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.unmap_lun(path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-unmap',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_unmap_lun_with_api_error(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        self.connection.invoke_successfully.side_effect =\
            netapp_api.NaApiError()

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.assertRaises(netapp_api.NaApiError, self.client.unmap_lun,
                              path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-unmap',
                **{'path': path, 'initiator-group': igroup})

    def test_unmap_lun_already_unmapped(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        EINVALIDINPUTERROR = '13115'
        self.connection.invoke_successfully.side_effect =\
            netapp_api.NaApiError(code=EINVALIDINPUTERROR)

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.unmap_lun(path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-unmap',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_unmap_lun_lun_not_mapped_in_group(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        igroup = 'igroup'
        EVDISK_ERROR_NO_SUCH_LUNMAP = '9016'
        self.connection.invoke_successfully.side_effect =\
            netapp_api.NaApiError(code=EVDISK_ERROR_NO_SUCH_LUNMAP)

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.unmap_lun(path, igroup)

            mock_create_node.assert_called_once_with(
                'lun-unmap',
                **{'path': path, 'initiator-group': igroup})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_create_igroup(self):
        igroup = 'igroup'

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.create_igroup(igroup)

            mock_create_node.assert_called_once_with(
                'igroup-create',
                **{'initiator-group-name': igroup,
                   'initiator-group-type': 'iscsi',
                   'os-type': 'default'})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_add_igroup_initiator(self):
        igroup = 'igroup'
        initiator = 'initiator'

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               ) as mock_create_node:
            self.client.add_igroup_initiator(igroup, initiator)

            mock_create_node.assert_called_once_with(
                'igroup-add',
                **{'initiator-group-name': igroup,
                   'initiator': initiator})
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_do_direct_resize(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        new_size = 1024
        mock_request = mock.Mock()

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               return_value=mock_request
                               ) as mock_create_node:
            self.client.do_direct_resize(path, new_size)

            mock_create_node.assert_called_once_with(
                'lun-resize',
                **{'path': path,
                   'size': new_size})
            mock_request.add_new_child.assert_called_once_with(
                'force', 'true')
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_do_direct_resize_not_forced(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        new_size = 1024
        mock_request = mock.Mock()

        with mock.patch.object(netapp_api.NaElement,
                               'create_node_with_children',
                               return_value=mock_request
                               ) as mock_create_node:
            self.client.do_direct_resize(path, new_size, force=False)

            mock_create_node.assert_called_once_with(
                'lun-resize',
                **{'path': path,
                   'size': new_size})
            self.assertFalse(mock_request.add_new_child.called)
            self.connection.invoke_successfully.assert_called_once_with(
                mock.ANY, True)

    def test_get_lun_geometry(self):
        expected_keys = set(['size', 'bytes_per_sector', 'sectors_per_track',
                             'tracks_per_cylinder', 'cylinders', 'max_resize'])
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        mock_response = mock.Mock()
        self.connection.invoke_successfully.return_value = mock_response

        geometry = self.client.get_lun_geometry(path)

        self.assertEqual(expected_keys, set(geometry.keys()))

    def test_get_lun_geometry_with_api_error(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        self.connection.invoke_successfully.side_effect =\
            netapp_api.NaApiError()
        geometry = self.client.get_lun_geometry(path)

        self.assertEqual({}, geometry)

    def test_get_volume_options(self):
        fake_response = netapp_api.NaElement('volume')
        fake_response.add_node_with_children('options', test='blah')
        self.connection.invoke_successfully.return_value = fake_response

        options = self.client.get_volume_options('volume')

        self.assertEqual(1, len(options))

    def test_get_volume_options_with_no_options(self):
        fake_response = netapp_api.NaElement('options')
        self.connection.invoke_successfully.return_value = fake_response

        options = self.client.get_volume_options('volume')

        self.assertEqual([], options)

    def test_move_lun(self):
        path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        new_path = '/vol/%s/%s' % (self.fake_volume, self.fake_lun)
        fake_response = netapp_api.NaElement('options')
        self.connection.invoke_successfully.return_value = fake_response

        self.client.move_lun(path, new_path)

        self.connection.invoke_successfully.assert_called_once_with(
            mock.ANY, True)

    def test_get_igroup_by_initiators(self):
        self.assertRaises(NotImplementedError,
                          self.client.get_igroup_by_initiators,
                          fake.FC_FORMATTED_INITIATORS)

    def test_get_fc_target_wwpns(self):
        self.assertRaises(NotImplementedError,
                          self.client.get_fc_target_wwpns)

    def test_has_luns_mapped_to_initiator(self):
        initiator = fake.FC_FORMATTED_INITIATORS[0]
        version_response = netapp_api.NaElement(
            etree.XML("""
  <results status="passed">
    <lun-maps>
      <lun-map-info>
        <path>/vol/cinder1/volume-9be956b3-9854-4a5c-a7f5-13a16da52c9c</path>
        <initiator-group>openstack-4b57a80b-ebca-4d27-bd63-48ac5408d08b
        </initiator-group>
        <lun-id>0</lun-id>
      </lun-map-info>
      <lun-map-info>
        <path>/vol/cinder1/volume-ac90433c-a560-41b3-9357-7f3f80071eb5</path>
        <initiator-group>openstack-4b57a80b-ebca-4d27-bd63-48ac5408d08b
        </initiator-group>
        <lun-id>1</lun-id>
      </lun-map-info>
    </lun-maps>
  </results>"""))

        self.connection.invoke_successfully.return_value = version_response

        self.assertTrue(self.client._has_luns_mapped_to_initiator(initiator))

    def test_has_luns_mapped_to_initiator_not_mapped(self):
        initiator = fake.FC_FORMATTED_INITIATORS[0]
        version_response = netapp_api.NaElement(
            etree.XML("""
  <results status="passed">
    <lun-maps />
  </results>"""))
        self.connection.invoke_successfully.return_value = version_response
        self.assertFalse(self.client._has_luns_mapped_to_initiator(initiator))

    @mock.patch.object(client_base.Client, '_has_luns_mapped_to_initiator')
    def test_has_luns_mapped_to_initiators(self,
                                           mock_has_luns_mapped_to_initiator):
        initiators = fake.FC_FORMATTED_INITIATORS
        mock_has_luns_mapped_to_initiator.return_value = True
        self.assertTrue(self.client.has_luns_mapped_to_initiators(initiators))

    @mock.patch.object(client_base.Client, '_has_luns_mapped_to_initiator')
    def test_has_luns_mapped_to_initiators_not_mapped(
            self, mock_has_luns_mapped_to_initiator):
        initiators = fake.FC_FORMATTED_INITIATORS
        mock_has_luns_mapped_to_initiator.return_value = False
        self.assertFalse(self.client.has_luns_mapped_to_initiators(initiators))

    def test_get_performance_counter_info(self):

        self.mock_send_request.return_value = netapp_api.NaElement(
            fake_client.PERF_OBJECT_COUNTER_LIST_INFO_WAFL_RESPONSE)

        result = self.client.get_performance_counter_info('wafl',
                                                          'cp_phase_times')

        expected = {
            'name': 'cp_phase_times',
            'base-counter': 'total_cp_msecs',
            'labels': fake_client.PERF_OBJECT_COUNTER_TOTAL_CP_MSECS_LABELS,
        }
        self.assertEqual(expected, result)

        perf_object_counter_list_info_args = {'objectname': 'wafl'}
        self.mock_send_request.assert_called_once_with(
            'perf-object-counter-list-info',
            perf_object_counter_list_info_args, enable_tunneling=False)

    def test_get_performance_counter_info_not_found(self):

        self.mock_send_request.return_value = netapp_api.NaElement(
            fake_client.PERF_OBJECT_COUNTER_LIST_INFO_WAFL_RESPONSE)

        self.assertRaises(exception.NotFound,
                          self.client.get_performance_counter_info,
                          'wafl',
                          'invalid')

    def test_delete_snapshot(self):
        api_args = {
            'volume': fake.SNAPSHOT['volume_id'],
            'snapshot': fake.SNAPSHOT['name'],
        }
        self.mock_object(self.client.connection, 'send_request')

        self.client.delete_snapshot(api_args['volume'],
                                    api_args['snapshot'])

        asserted_api_args = {
            'volume': api_args['volume'],
            'snapshot': api_args['snapshot'],
        }
        self.client.connection.send_request.assert_called_once_with(
            'snapshot-delete', asserted_api_args)

    def test_create_cg_snapshot(self):
        self.mock_object(self.client, '_start_cg_snapshot',
                         return_value=fake.CONSISTENCY_GROUP_ID)
        self.mock_object(self.client, '_commit_cg_snapshot')

        self.client.create_cg_snapshot([fake.CG_VOLUME_NAME],
                                       fake.CG_SNAPSHOT_NAME)

        self.client._commit_cg_snapshot.assert_called_once_with(
            fake.CONSISTENCY_GROUP_ID)

    def test_create_cg_snapshot_no_id(self):
        self.mock_object(self.client, '_start_cg_snapshot', return_value=None)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.client.create_cg_snapshot,
                          [fake.CG_VOLUME_NAME],
                          fake.CG_SNAPSHOT_NAME)

    def test_start_cg_snapshot(self):
        snapshot_init = {
            'snapshot': fake.CG_SNAPSHOT_NAME,
            'timeout': 'relaxed',
            'volumes': [{'volume-name': fake.CG_VOLUME_NAME}],
        }
        self.mock_object(self.client.connection, 'send_request')

        self.client._start_cg_snapshot([fake.CG_VOLUME_NAME],
                                       snapshot_init['snapshot'])

        self.client.connection.send_request.assert_called_once_with(
            'cg-start', snapshot_init)

    def test_commit_cg_snapshot(self):
        snapshot_commit = {'cg-id': fake.CG_VOLUME_ID}
        self.mock_object(self.client.connection, 'send_request')

        self.client._commit_cg_snapshot(snapshot_commit['cg-id'])

        self.client.connection.send_request.assert_called_once_with(
            'cg-commit', {'cg-id': snapshot_commit['cg-id']})

    def test_wait_for_busy_snapshot_raise_exception(self):
        BUSY_SNAPSHOT = dict(fake.SNAPSHOT)
        BUSY_SNAPSHOT['busy'] = True

        # Need to mock sleep as it is called by @utils.retry
        self.mock_object(time, 'sleep')
        mock_get_snapshot = self.mock_object(self.client, 'get_snapshot',
                                             return_value=BUSY_SNAPSHOT)

        self.assertRaises(exception.SnapshotIsBusy,
                          self.client.wait_for_busy_snapshot,
                          fake.FLEXVOL, fake.SNAPSHOT_NAME)

        calls = [
            mock.call(fake.FLEXVOL, fake.SNAPSHOT_NAME),
            mock.call(fake.FLEXVOL, fake.SNAPSHOT_NAME),
            mock.call(fake.FLEXVOL, fake.SNAPSHOT_NAME),
        ]
        mock_get_snapshot.assert_has_calls(calls)

    def test_rename_snapshot(self):
        self.mock_object(self.client.connection, 'send_request')

        self.client.rename_snapshot(
            fake.SNAPSHOT['volume_id'], fake.SNAPSHOT_NAME,
            client_base.DELETED_PREFIX + fake.SNAPSHOT_NAME)

        api_args = {
            'volume': fake.SNAPSHOT['volume_id'],
            'current-name': fake.SNAPSHOT_NAME,
            'new-name':
                client_base.DELETED_PREFIX + fake.SNAPSHOT_NAME,
        }

        self.client.connection.send_request.assert_called_once_with(
            'snapshot-rename', api_args)



# Copyright 2020 Thomas Goirand <zigo@debian.org>
# Copyright 2020 Infomaniak Networks.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
import requests

from cinder.tests.functional import functional_helpers


class TestHealthCheckMiddleware(functional_helpers._FunctionalTestBase):

    def test_healthcheck(self):
        # We verify that we return a HTTP200 when calling api_get
        url = 'http://%s:%s/healthcheck' % (self.osapi.host, self.osapi.port)
        response = requests.request(
            'GET',
            url,
            headers={'Accept': 'application/json'})
        output = jsonutils.loads(response.content)
        self.assertEqual(['OK'], output['reasons'])



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for our unit tests.

Allows overriding of CONF for use of fakes, and some black magic for
inline callbacks.

"""
import copy
import logging
import os
from unittest import mock
import uuid

from eventlet import tpool
import fixtures
from keystonemiddleware import auth_token
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_log.fixture import logging_error as log_fixture
import oslo_messaging
from oslo_messaging import conffixture as messaging_conffixture
from oslo_serialization import jsonutils
from oslo_utils import strutils
from oslo_utils import timeutils
import six
import testtools

from cinder.api import common as api_common
from cinder.common import config
from cinder import context
from cinder import coordination
from cinder.db import migration
from cinder.db.sqlalchemy import api as sqla_api
from cinder import i18n
from cinder.objects import base as objects_base
from cinder import rpc
from cinder import service
from cinder.tests import fixtures as cinder_fixtures
from cinder.tests.unit import conf_fixture
from cinder.tests.unit import fake_notifier
from cinder.volume import configuration
from cinder.volume import driver as vol_driver
from cinder.volume import volume_types
from cinder.volume import volume_utils

CONF = config.CONF

_DB_CACHE = None
SESSION_CONFIGURED = False


class TestingException(Exception):
    pass


class Database(fixtures.Fixture):

    def __init__(self, db_api, db_migrate, sql_connection):
        # NOTE(lhx_): oslo_db.enginefacade is configured in tests the same
        # way as it's done for any other services that uses the db
        global SESSION_CONFIGURED
        if not SESSION_CONFIGURED:
            sqla_api.configure(CONF)
            SESSION_CONFIGURED = True
        self.sql_connection = sql_connection

        # Suppress logging for test runs
        migrate_logger = logging.getLogger('migrate')
        migrate_logger.setLevel(logging.WARNING)

        self.engine = db_api.get_engine()
        self.engine.dispose()
        conn = self.engine.connect()
        db_migrate.db_sync()
        self._DB = "".join(line for line in conn.connection.iterdump())
        self.engine.dispose()

    def setUp(self):
        super(Database, self).setUp()

        conn = self.engine.connect()
        conn.connection.executescript(self._DB)
        self.addCleanup(self.engine.dispose)


class TestCase(testtools.TestCase):
    """Test case base class for all unit tests."""

    SOURCE_TREE_ROOT = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '../../../'))
    POLICY_PATH = os.path.join(SOURCE_TREE_ROOT,
                               'cinder/tests/unit/policy.json')
    RESOURCE_FILTER_FILENAME = 'etc/cinder/resource_filters.json'
    RESOURCE_FILTER_PATH = os.path.join(SOURCE_TREE_ROOT,
                                        RESOURCE_FILTER_FILENAME)
    MOCK_WORKER = True
    MOCK_TOOZ = True

    def __init__(self, *args, **kwargs):
        super(TestCase, self).__init__(*args, **kwargs)

        # Suppress some log messages during test runs
        castellan_logger = logging.getLogger('castellan')
        castellan_logger.setLevel(logging.ERROR)
        stevedore_logger = logging.getLogger('stevedore')
        stevedore_logger.setLevel(logging.ERROR)

    def _get_joined_notifier(self, *args, **kwargs):
        # We create a new fake notifier but we join the notifications with
        # the default notifier
        notifier = fake_notifier.get_fake_notifier(*args, **kwargs)
        notifier.notifications = self.notifier.notifications
        return notifier

    def _reset_filter_file(self):
        self.override_config('resource_query_filters_file',
                             self.RESOURCE_FILTER_PATH)
        api_common._FILTERS_COLLECTION = None

    def setUp(self):
        """Run before each test method to initialize test environment."""
        super(TestCase, self).setUp()

        # Create default notifier
        self.notifier = fake_notifier.get_fake_notifier()

        # Mock rpc get notifier with fake notifier method that joins all
        # notifications with the default notifier
        self.patch('cinder.rpc.get_notifier',
                   side_effect=self._get_joined_notifier)

        # Protect against any case where someone doesn't directly patch a retry
        # decorated call.
        self.patch('tenacity.nap.sleep')

        if self.MOCK_WORKER:
            # Mock worker creation for all tests that don't care about it
            clean_path = 'cinder.objects.cleanable.CinderCleanableObject.%s'
            for method in ('create_worker', 'set_worker', 'unset_worker'):
                self.patch(clean_path % method, return_value=None)

        if self.MOCK_TOOZ:
            self.patch('cinder.coordination.Coordinator.start')
            self.patch('cinder.coordination.Coordinator.stop')
            self.patch('cinder.coordination.Coordinator.get_lock')

        # Unit tests do not need to use lazy gettext
        i18n.enable_lazy(False)

        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 0)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            # If timeout value is invalid do not set a timeout.
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())

        environ_enabled = (lambda var_name:
                           strutils.bool_from_string(os.environ.get(var_name)))
        if environ_enabled('OS_STDOUT_CAPTURE'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if environ_enabled('OS_STDERR_CAPTURE'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(cinder_fixtures.StandardLogging())

        rpc.add_extra_exmods("cinder.tests.unit")
        self.addCleanup(rpc.clear_extra_exmods)
        self.addCleanup(rpc.cleanup)

        self.messaging_conf = messaging_conffixture.ConfFixture(CONF)
        self.messaging_conf.transport_url = 'fake:/'
        self.messaging_conf.response_timeout = 15
        self.useFixture(self.messaging_conf)

        # Load oslo_messaging_notifications config group so we can set an
        # override to prevent notifications from being ignored due to the
        # short-circuit mechanism.
        oslo_messaging.get_notification_transport(CONF)
        #  We need to use a valid driver for the notifications, so we use test.
        self.override_config('driver', ['test'],
                             group='oslo_messaging_notifications')
        rpc.init(CONF)

        # NOTE(geguileo): This is required because _determine_obj_version_cap
        # and _determine_rpc_version_cap functions in cinder.rpc.RPCAPI cache
        # versions in LAST_RPC_VERSIONS and LAST_OBJ_VERSIONS so we may have
        # weird interactions between tests if we don't clear them before each
        # test.
        rpc.LAST_OBJ_VERSIONS = {}
        rpc.LAST_RPC_VERSIONS = {}

        # Init AuthProtocol to register some base options first, such as
        # auth_url.
        auth_token.AuthProtocol('fake_app', {'auth_type': 'password',
                                             'auth_url': 'fake_url'})

        conf_fixture.set_defaults(CONF)
        CONF([], default_config_files=[])

        # NOTE(vish): We need a better method for creating fixtures for tests
        #             now that we have some required db setup for the system
        #             to work properly.
        self.start = timeutils.utcnow()

        CONF.set_default('connection', 'sqlite://', 'database')
        CONF.set_default('sqlite_synchronous', False, 'database')

        global _DB_CACHE
        if not _DB_CACHE:
            _DB_CACHE = Database(sqla_api, migration,
                                 sql_connection=CONF.database.connection)
        self.useFixture(_DB_CACHE)

        # NOTE(blk-u): WarningsFixture must be after the Database fixture
        # because sqlalchemy-migrate messes with the warnings filters.
        self.useFixture(cinder_fixtures.WarningsFixture())

        # NOTE(danms): Make sure to reset us back to non-remote objects
        # for each test to avoid interactions. Also, backup the object
        # registry.
        objects_base.CinderObject.indirection_api = None
        self._base_test_obj_backup = copy.copy(
            objects_base.CinderObjectRegistry._registry._obj_classes)
        self.addCleanup(self._restore_obj_registry)

        self.addCleanup(CONF.reset)
        self.addCleanup(self._common_cleanup)
        self._services = []

        fake_notifier.mock_notifier(self)

        # This will be cleaned up by the NestedTempfile fixture
        lock_path = self.useFixture(fixtures.TempDir()).path
        self.fixture = self.useFixture(
            config_fixture.Config(lockutils.CONF))
        self.fixture.config(lock_path=lock_path,
                            group='oslo_concurrency')
        lockutils.set_defaults(lock_path)
        self.override_config('policy_file',
                             os.path.join(
                                 os.path.abspath(
                                     os.path.dirname(__file__)
                                 ),
                                 self.POLICY_PATH),
                             group='oslo_policy')
        self.override_config('resource_query_filters_file',
                             self.RESOURCE_FILTER_PATH)
        self._disable_osprofiler()

        # NOTE(geguileo): This is required because common get_by_id method in
        # cinder.db.sqlalchemy.api caches get methods and if we use a mocked
        # get method in one test it would carry on to the next test.  So we
        # clear out the cache.
        sqla_api._GET_METHODS = {}

        self.override_config('backend_url', 'file://' + lock_path,
                             group='coordination')
        coordination.COORDINATOR.start()
        self.addCleanup(coordination.COORDINATOR.stop)

        if six.PY3:
            # TODO(smcginnis) Python 3 deprecates assertRaisesRegexp to
            # assertRaisesRegex, but Python 2 does not have the new name. This
            # can be removed once we stop supporting py2 or the new name is
            # added.
            self.assertRaisesRegexp = self.assertRaisesRegex

        # Ensure we have the default tpool size value and we don't carry
        # threads from other test runs.
        tpool.killall()
        tpool._nthreads = 20

        # NOTE(mikal): make sure we don't load a privsep helper accidentally
        self.useFixture(cinder_fixtures.PrivsepNoHelperFixture())

        # NOTE: This volume type is created to avoid failure at database since
        # volume_type_id is non-nullable for volumes and snapshots

        self.vt = volume_types.get_default_volume_type()

    def _restore_obj_registry(self):
        objects_base.CinderObjectRegistry._registry._obj_classes = \
            self._base_test_obj_backup

    def _disable_osprofiler(self):
        """Disable osprofiler.

        osprofiler should not run for unit tests.
        """

        def side_effect(value):
            return value
        mock_decorator = mock.MagicMock(side_effect=side_effect)
        p = mock.patch("osprofiler.profiler.trace_cls",
                       return_value=mock_decorator)
        p.start()

    def _common_cleanup(self):
        """Runs after each test method to tear down test environment."""

        # Kill any services
        for x in self._services:
            try:
                x.kill()
            except Exception:
                pass

        # Delete attributes that don't start with _ so they don't pin
        # memory around unnecessarily for the duration of the test
        # suite
        for key in [k for k in self.__dict__ if k[0] != '_']:
            del self.__dict__[key]

    def override_config(self, name, override, group=None):
        """Cleanly override CONF variables."""
        CONF.set_override(name, override, group)
        self.addCleanup(CONF.clear_override, name, group)

    def flags(self, **kw):
        """Override CONF variables for a test."""
        group = kw.pop('group', None)
        for k, v in kw.items():
            self.override_config(k, v, group)

    def start_service(self, name, host=None, **kwargs):
        host = host if host else uuid.uuid4().hex
        kwargs.setdefault('host', host)
        kwargs.setdefault('binary', 'cinder-%s' % name)
        svc = service.Service.create(**kwargs)
        svc.start()
        self._services.append(svc)
        return svc

    def mock_object(self, obj, attr_name, *args, **kwargs):
        """Use python mock to mock an object attribute

        Mocks the specified objects attribute with the given value.
        Automatically performs 'addCleanup' for the mock.

        """
        patcher = mock.patch.object(obj, attr_name, *args, **kwargs)
        result = patcher.start()
        self.addCleanup(patcher.stop)
        return result

    def patch(self, path, *args, **kwargs):
        """Use python mock to mock a path with automatic cleanup."""
        patcher = mock.patch(path, *args, **kwargs)
        result = patcher.start()
        self.addCleanup(patcher.stop)
        return result

    # Useful assertions
    def assert_notify_called(self, mock_notify, calls, any_order=False):
        if any_order is True:
            for c in calls:
                # mock_notify.call_args_list = [
                #     mock.call('INFO', 'volume.retype', ...),
                #     mock.call('WARN', 'cinder.fire', ...)]
                # m = mock_notify.call_args_list
                # m[0] = Call
                # m[0][0] = tuple('INFO', <context>, 'volume.retype', ...)
                if not any(m for m in mock_notify.call_args_list
                           if (m[0][0] == c[0]     # 'INFO'
                               and
                               m[0][2] == c[1])):  # 'volume.retype'
                    raise AssertionError("notify call not found: %s" % c)
            return

        for i in range(0, len(calls)):
            mock_call = mock_notify.call_args_list[i]
            call = calls[i]

            posargs = mock_call[0]

            self.assertEqual(call[0], posargs[0])
            self.assertEqual(call[1], posargs[2])

    def assertTrue(self, x, *args, **kwargs):
        """Assert that value is True.

        If original behavior is required we will need to do:
            assertTrue(bool(result))
        """
        # assertTrue uses msg but assertIs uses message keyword argument
        args = list(args)
        msg = kwargs.pop('msg', args.pop(0) if args else '')
        kwargs.setdefault('message', msg)
        self.assertIs(True, x, *args, **kwargs)

    def assertFalse(self, x, *args, **kwargs):
        """Assert that value is False.

        If original behavior is required we will need to do:
            assertFalse(bool(result))
        """
        # assertTrue uses msg but assertIs uses message keyword argument
        args = list(args)
        msg = kwargs.pop('msg', args.pop(0) if args else '')
        kwargs.setdefault('message', msg)
        self.assertIs(False, x, *args, **kwargs)

    def stub_out(self, old, new):
        """Replace a function for the duration of the test.

        Use the monkey patch fixture to replace a function for the
        duration of a test. Useful when you want to provide fake
        methods instead of mocks during testing.
        This should be used instead of self.stubs.Set (which is based
        on mox) going forward.
        """
        self.useFixture(fixtures.MonkeyPatch(old, new))

    def _set_unique_fqdn_override(self, value, in_shared):
        """Override the unique_fqdn_network configuration option.

        Meant for driver tests that use a Mock for their driver configuration
        instead of a real Oslo Conf.
        """
        # Since we don't use a real oslo config for the driver we don't get
        # the default initialization, so create a group and register the option
        cfg.CONF.register_group(cfg.OptGroup('driver_cfg'))
        new_config = configuration.Configuration([], config_group='driver_cfg')
        new_config.append_config_values(vol_driver.fqdn_opts)

        # Now we override the value for this test
        group = configuration.SHARED_CONF_GROUP if in_shared else 'driver_cfg'
        self.addCleanup(CONF.clear_override, 'unique_fqdn_network',
                        group=group)
        cfg.CONF.set_override('unique_fqdn_network', value, group=group)
        return new_config


class ModelsObjectComparatorMixin(object):
    def _dict_from_object(self, obj, ignored_keys):
        if ignored_keys is None:
            ignored_keys = []
        obj = jsonutils.to_primitive(obj)  # Convert to dict first.
        items = obj.items()
        return {k: v for k, v in items
                if k not in ignored_keys}

    def _assertEqualObjects(self, obj1, obj2, ignored_keys=None):
        obj1 = self._dict_from_object(obj1, ignored_keys)
        obj2 = self._dict_from_object(obj2, ignored_keys)

        self.assertEqual(
            len(obj1), len(obj2),
            "Keys mismatch: %s" % six.text_type(
                set(obj1.keys()) ^ set(obj2.keys())))
        for key, value in obj1.items():
            self.assertEqual(value, obj2[key])

    def _assertEqualListsOfObjects(self, objs1, objs2, ignored_keys=None,
                                   msg=None):
        def obj_to_dict(o):
            return self._dict_from_object(o, ignored_keys)

        objs1 = map(obj_to_dict, objs1)
        objs2 = list(map(obj_to_dict, objs2))
        # We don't care about the order of the lists, as long as they are in
        for obj1 in objs1:
            self.assertIn(obj1, objs2)
            objs2.remove(obj1)
        self.assertEqual([], objs2)

    def _assertEqualListsOfPrimitivesAsSets(self, primitives1, primitives2):
        self.assertEqual(len(primitives1), len(primitives2))
        for primitive in primitives1:
            self.assertIn(primitive, primitives2)

        for primitive in primitives2:
            self.assertIn(primitive, primitives1)


class RPCAPITestCase(TestCase, ModelsObjectComparatorMixin):
    def setUp(self):
        super(RPCAPITestCase, self).setUp()
        self.context = context.get_admin_context()
        self.rpcapi = None
        self.base_version = '2.0'

    def _test_rpc_api(self, method, rpc_method, server=None, fanout=False,
                      version=None, expected_method=None,
                      expected_kwargs_diff=None, retval=None,
                      expected_retval=None, **kwargs):
        """Runs a test against RPC API method.

        :param method: Name of RPC API method.
        :param rpc_method: Expected RPC message type (cast or call).
        :param server: Expected hostname.
        :param fanout: True if expected call/cast should be fanout.
        :param version: Expected autocalculated RPC API version.
        :param expected_method: Expected RPC method name.
        :param expected_kwargs_diff: Map of expected changes between keyword
                                     arguments passed into the method and sent
                                     over RPC.
        :param retval: Value returned by RPC call/cast.
        :param expected_retval: Expected RPC API response (if different than
                                retval).
        :param kwargs: Parameters passed into the RPC API method.
        """

        rpcapi = self.rpcapi()
        expected_kwargs_diff = expected_kwargs_diff or {}
        version = version or self.base_version
        topic = None
        if server is not None:
            backend = volume_utils.extract_host(server)
            server = volume_utils.extract_host(server, 'host')
            topic = 'cinder-volume.%s' % backend

        if expected_method is None:
            expected_method = method

        if expected_retval is None:
            expected_retval = retval

        target = {
            "server": server,
            "fanout": fanout,
            "version": version,
            "topic": topic,
        }

        # Initially we expect that we'll pass same arguments to RPC API method
        # and RPC call/cast...
        expected_msg = copy.deepcopy(kwargs)
        # ... but here we're taking exceptions into account.
        expected_msg.update(expected_kwargs_diff)

        def _fake_prepare_method(*args, **kwds):
            # This is checking if target will be properly created.
            for kwd in kwds:
                self.assertEqual(target[kwd], kwds[kwd])
            return rpcapi.client

        def _fake_rpc_method(*args, **kwargs):
            # This checks if positional arguments passed to RPC method match.
            self.assertEqual((self.context, expected_method), args)

            # This checks if keyword arguments passed to RPC method match.
            for kwarg, value in kwargs.items():
                # Getting possible changes into account.
                if isinstance(value, objects_base.CinderObject):
                    # We need to compare objects differently.
                    self._assertEqualObjects(expected_msg[kwarg], value)
                else:
                    self.assertEqual(expected_msg[kwarg], value)

            # Returning fake value we're supposed to return.
            if retval:
                return retval

        # Enable mocks that will check everything and run RPC method.
        with mock.patch.object(rpcapi.client, "prepare",
                               side_effect=_fake_prepare_method):
            with mock.patch.object(rpcapi.client, rpc_method,
                                   side_effect=_fake_rpc_method):
                real_retval = getattr(rpcapi, method)(self.context, **kwargs)
                self.assertEqual(expected_retval, real_retval)



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common
from cinder.api import microversions as mv
from cinder.policies import group_snapshots as policy


class ViewBuilder(common.ViewBuilder):
    """Model group_snapshot API responses as a python dictionary."""

    _collection_name = "group_snapshots"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, group_snapshots):
        """Show a list of group_snapshots without many details."""
        return self._list_view(self.summary, request, group_snapshots)

    def detail_list(self, request, group_snapshots):
        """Detailed view of a list of group_snapshots ."""
        return self._list_view(self.detail, request, group_snapshots)

    def summary(self, request, group_snapshot):
        """Generic, non-detailed view of a group_snapshot."""
        return {
            'group_snapshot': {
                'id': group_snapshot.id,
                'name': group_snapshot.name,
                # NOTE(xyang): group_type_id is added for migrating CGs
                # to generic volume groups
                'group_type_id': group_snapshot.group_type_id,
            }
        }

    def detail(self, request, group_snapshot):
        """Detailed view of a single group_snapshot."""
        group_snapshot_ref = {
            'group_snapshot': {
                'id': group_snapshot.id,
                'group_id': group_snapshot.group_id,
                'group_type_id': group_snapshot.group_type_id,
                'status': group_snapshot.status,
                'created_at': group_snapshot.created_at,
                'name': group_snapshot.name,
                'description': group_snapshot.description
            }
        }
        req_version = request.api_version_request
        context = request.environ['cinder.context']

        if req_version.matches(mv.GROUP_GROUPSNAPSHOT_PROJECT_ID, None):
            if context.authorize(policy.GROUP_SNAPSHOT_ATTRIBUTES_POLICY,
                                 fatal=False):
                group_snapshot_ref['group_snapshot']['project_id'] = (
                    group_snapshot.project_id)

        return group_snapshot_ref

    def _list_view(self, func, request, group_snapshots):
        """Provide a view for a list of group_snapshots."""
        group_snapshots_list = [func(request, group_snapshot)['group_snapshot']
                                for group_snapshot in group_snapshots]
        group_snapshot_links = self._get_collection_links(
            request, group_snapshots_list, self._collection_name)
        group_snapshots_dict = dict(group_snapshots=group_snapshots_list)
        if group_snapshot_links:
            group_snapshots_dict['group_snapshot_links'] = group_snapshot_links

        return group_snapshots_dict



# Copyright 2017 Veritas Technologies LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Veritas Access Driver for ISCSI.

"""
import ast
import hashlib
import json
from random import randint
from xml.dom import minidom

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import netutils
from oslo_utils import strutils
from oslo_utils import units
import requests
import requests.auth
from six.moves import http_client

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.san import san

LOG = logging.getLogger(__name__)


VA_VOL_OPTS = [
    cfg.BoolOpt('vrts_lun_sparse',
                default=True,
                help='Create sparse Lun.'),
    cfg.StrOpt('vrts_target_config',
               default='/etc/cinder/vrts_target.xml',
               help='VA config file.')
]


CONF = cfg.CONF
CONF.register_opts(VA_VOL_OPTS)


class NoAuth(requests.auth.AuthBase):
    """This is a 'authentication' handler.

    It exists for use with custom authentication systems, such as the
    one for the Access API, it simply passes the Authorization header as-is.

    The default authentication handler for requests will clobber the
    Authorization header.
    """

    def __call__(self, r):
        return r


@interface.volumedriver
class ACCESSIscsiDriver(driver.ISCSIDriver):
    """ACCESS Share Driver.

    Executes commands relating to ACCESS ISCSI.
    Supports creation of volumes on ACCESS.

    .. code-block:: none

     API version history:

        1.0 - Initial version.
    """

    VERSION = "1.0"
    # ThirdPartySytems wiki page
    CI_WIKI_NAME = "Veritas_Access_CI"
    DRIVER_VOLUME_TYPE = 'iSCSI'
    LUN_FOUND_INTERVAL = 30  # seconds

    # TODO(jsbryant) Remove driver in the 'U' release if CI is not fixed.
    SUPPORTED = False

    def __init__(self, *args, **kwargs):
        # Parent sets db, host, _execute and base config
        super(ACCESSIscsiDriver, self).__init__(*args, **kwargs)

        self._va_ip = None
        self._port = None
        self._user = None
        self._pwd = None
        self.iscsi_port = None
        self._fs_list_str = '/fs'
        self._target_list_str = '/iscsi/target/list'
        self._target_status = '/iscsi/target/status'
        self._lun_create_str = '/iscsi/lun/create'
        self._lun_destroy_str = '/iscsi/lun/destroy'
        self._lun_list_str = '/iscsi/lun/list'
        self._lun_create_from_snap_str = '/iscsi/lun_from_snap/create'
        self._snapshot_create_str = '/iscsi/lun/snapshot/create'
        self._snapshot_destroy_str = '/iscsi/lun/snapshot/destroy'
        self._snapshot_list_str = '/iscsi/lun/snapshot/list'
        self._lun_clone_create_str = '/iscsi/lun/clone/create'
        self._lun_extend_str = '/iscsi/lun/growto'
        self._lun_shrink_str = '/iscsi/lun/shrinkto'
        self._lun_getid_str = '/iscsi/lun/getlunid'
        self._target_map_str = '/iscsi/target/map/add'
        self._target_list_status = '/iscsi/target/full_list'

        self.configuration.append_config_values(VA_VOL_OPTS)
        self.configuration.append_config_values(san.san_opts)
        self.backend_name = (self.configuration.safe_get('volume_'
                                                         'backend_name') or
                             'ACCESS_ISCSI')
        self.verify = (self.configuration.
                       safe_get('driver_ssl_cert_verify') or False)

        if self.verify:
            verify_path = (self.configuration.
                           safe_get('driver_ssl_cert_path') or None)
            if verify_path:
                self.verify = verify_path

    @staticmethod
    def get_driver_options():
        return VA_VOL_OPTS

    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        super(ACCESSIscsiDriver, self).do_setup(context)

        required_config = ['san_ip',
                           'san_login',
                           'san_password',
                           'san_api_port']

        for attr in required_config:
            if not getattr(self.configuration, attr, None):
                message = (_('config option %s is not set.') % attr)
                raise exception.InvalidInput(message=message)

        self._va_ip = self.configuration.san_ip
        self._user = self.configuration.san_login
        self._pwd = self.configuration.san_password
        self._port = self.configuration.san_api_port
        self._sparse_lun_support = self.configuration.vrts_lun_sparse
        self.target_info_file = self.configuration.vrts_target_config
        self.iscsi_port = self.configuration.target_port
        self.session = self._authenticate_access(self._va_ip, self._user,
                                                 self._pwd)

    def _get_va_lun_name(self, name):
        length = len(name)
        index = int(length / 2)
        name1 = name[:index]
        name2 = name[index:]
        crc1 = hashlib.md5(name1.encode('utf-8')).hexdigest()[:5]
        crc2 = hashlib.md5(name2.encode('utf-8')).hexdigest()[:5]
        return 'cinder' + '-' + crc1 + '-' + crc2

    def check_for_setup_error(self):
        """Check if veritas access target is online."""
        target_list = self._vrts_parse_xml_file(self.target_info_file)
        if not self._vrts_get_online_targets(target_list):
            message = ('ACCESSIscsiDriver setup error as '
                       'no target is online')
            raise exception.VolumeBackendAPIException(message=message)

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        pass

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume."""
        pass

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""
        pass

    def _vrts_get_iscsi_properties(self, volume, target_name):
        """Get target and LUN details."""
        lun_name = self._get_va_lun_name(volume.id)

        data = {}
        path = self._lun_getid_str
        provider = '%s:%s' % (self._va_ip, self._port)

        lun_id_list = self._access_api(self.session, provider, path,
                                       json.dumps(data), 'GET')

        if not lun_id_list:
            message = _('ACCESSIscsiDriver get LUN ID list '
                        'operation failed')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        for lun in ast.literal_eval(lun_id_list['output']):
            vrts_lun_name = lun['storage_object'].split('/')[3]
            if vrts_lun_name == lun_name:
                lun_id = int(lun['index'])

        target_list = self._vrts_parse_xml_file(self.target_info_file)
        authentication = False
        portal_ip = ""

        for target in target_list:
            if target_name == target['name']:
                portal_ip = target['portal_ip']
                if target['auth'] == '1':
                    auth_user = target['auth_user']
                    auth_password = target['auth_password']
                    authentication = True
                break

        if portal_ip == "":
            message = (_('ACCESSIscsiDriver initialize_connection '
                         'failed for %s as no portal ip was found')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        portal_list = portal_ip.split(',')

        target_portal_list = []
        for ip in portal_list:
            if netutils.is_valid_ipv6(ip):
                target_portal_list.append('[%s]:%s' % (ip,
                                                       str(self.iscsi_port)))
            else:
                target_portal_list.append('%s:%s' % (ip, str(self.iscsi_port)))

        iscsi_properties = {}
        iscsi_properties['target_discovered'] = True
        iscsi_properties['target_iqn'] = target_name
        iscsi_properties['target_portal'] = target_portal_list[0]
        if len(target_portal_list) > 1:
            iscsi_properties['target_portals'] = target_portal_list
        iscsi_properties['target_lun'] = lun_id
        iscsi_properties['volume_id'] = volume.id
        if authentication:
            iscsi_properties['auth_username'] = auth_user
            iscsi_properties['auth_password'] = auth_password
            iscsi_properties['auth_method'] = 'CHAP'

        return iscsi_properties

    def _get_vrts_lun_list(self):
        """Get Lun list."""
        data = {}
        path = self._lun_list_str
        provider = '%s:%s' % (self._va_ip, self._port)

        lun_list = self._access_api(self.session, provider, path,
                                    json.dumps(data), 'GET')

        if not lun_list:
            message = _('ACCESSIscsiDriver get LUN list '
                        'operation failed')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        return lun_list

    def _vrts_target_initiator_mapping(self, target_name, initiator_name):
        """Map target to initiator."""
        path = self._target_map_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["target_name"] = target_name
        data["initiator_name"] = initiator_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver target-initiator mapping '
                         'failed for target %s')
                       % target_name)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def initialize_connection(self, volume, connector, initiator_data=None):
        """Initializes the connection and returns connection info.

        The iscsi driver returns a driver_volume_type of 'iscsi'.
        the format of the driver data is defined in _vrts_get_iscsi_properties.
        Example return value::

            {
                'driver_volume_type': 'iscsi'
                'data': {
                    'target_discovered': True,
                    'target_iqn': 'iqn.2010-10.org.openstack:volume-00000001',
                    'target_portal': '127.0.0.0.1:3260',
                    'target_lun': 1,
                    'volume_id': '12345678-1234-4321-1234-123456789012',
                }
            }
        """
        lun_name = self._get_va_lun_name(volume.id)
        target = {'target_name': ''}

        def _inner():
            lun_list = self._get_vrts_lun_list()
            for lun in lun_list['output']['output']['luns']:
                if lun['lun_name'] == lun_name:
                    target['target_name'] = lun['target_name']
                    raise loopingcall.LoopingCallDone()

        timer = loopingcall.FixedIntervalWithTimeoutLoopingCall(_inner)
        try:
            timer.start(interval=5, timeout=self.LUN_FOUND_INTERVAL).wait()
        except loopingcall.LoopingCallTimeOut:
            message = (_('ACCESSIscsiDriver initialize_connection '
                         'failed for %s as no target was found')
                       % volume.id)

            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        self._vrts_target_initiator_mapping(target['target_name'],
                                            connector['initiator'])

        iscsi_properties = self._vrts_get_iscsi_properties(
            volume, target['target_name'])

        return {
            'driver_volume_type': 'iscsi',
            'data': iscsi_properties
        }

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        pass

    def _vrts_parse_xml_file(self, filename):
        """VRTS target info.

        <VRTS>
        <VrtsTargets>
            <Target>
                <Name>iqn.2017-02.com.veritas:target03</Name>
                <PortalIP>10.182.174.188</PortalIP>
            </Target>
            <Target>
                <Name>iqn.2017-02.com.veritas:target04</Name>
                <PortalIP>10.182.174.189</PortalIP>
            </Target>
        </VrtsTargets>
        </VRST>

        :param filename: the configuration file
        :returns: list
        """
        myfile = open(filename, 'r')
        data = myfile.read()
        myfile.close()
        dom = minidom.parseString(data)

        mylist = []
        target = {}

        try:
            for trg in dom.getElementsByTagName('Target'):
                target['name'] = (trg.getElementsByTagName('Name')[0]
                                  .childNodes[0].nodeValue)
                target['portal_ip'] = (trg.getElementsByTagName('PortalIP')[0]
                                       .childNodes[0].nodeValue)
                target['auth'] = (trg.getElementsByTagName('Authentication')[0]
                                  .childNodes[0].nodeValue)
                if target['auth'] == '1':
                    target['auth_user'] = (trg.getElementsByTagName
                                           ('Auth_username')[0]
                                           .childNodes[0].nodeValue)
                    target['auth_password'] = (trg.getElementsByTagName
                                               ('Auth_password')[0]
                                               .childNodes[0].nodeValue)

                mylist.append(target)
                target = {}
        except IndexError:
            pass

        return mylist

    def _vrts_get_fs_list(self):
        """Get FS list."""
        path = self._fs_list_str
        provider = '%s:%s' % (self._va_ip, self._port)
        data = {}
        fs_list = self._access_api(self.session, provider, path,
                                   json.dumps(data), 'GET')

        if not fs_list:
            message = _('ACCESSIscsiDriver get FS list '
                        'operation failed')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        return fs_list

    def _vrts_get_online_targets(self, available_targets):
        """Out of available targets get list of targets which are online."""

        online_targets = []
        path = self._target_list_status
        provider = '%s:%s' % (self._va_ip, self._port)
        data = {}
        target_status_list = self._access_api(self.session, provider, path,
                                              json.dumps(data), 'GET')

        try:
            target_status_output = (ast.
                                    literal_eval(target_status_list['output']))
        except KeyError:
            message = _('ACCESSIscsiDriver get online target list '
                        'operation failed')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        for target in available_targets:
            if target['name'] in target_status_output.keys():
                if target_status_output[target['name']] == 'ONLINE':
                    online_targets.append(target)

        return online_targets

    def _vrts_get_targets_store(self):
        """Get target and its store list."""
        path = self._target_list_str
        provider = '%s:%s' % (self._va_ip, self._port)
        data = {}
        target_list = self._access_api(self.session, provider, path,
                                       json.dumps(data), 'GET')

        if not target_list:
            message = _('ACCESSIscsiDriver get target list '
                        'operation failed')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        return target_list['output']['output']['targets']

    def _vrts_get_assigned_store(self, target, vrts_target_list):
        """Get the store mapped to given target."""
        for vrts_target in vrts_target_list:
            if vrts_target['wwn'] == target:
                return vrts_target['fs_list'][0]

    def _vrts_is_space_available_in_store(self, vol_size, store_name, fs_list):
        """Check whether space is available on store."""
        if self._sparse_lun_support:
            return True

        for fs in fs_list:
            if fs['name'] == store_name:
                fs_avilable_space = (int(fs['file_storage_capacity']) -
                                     int(fs['file_storage_used']))
                free_space = fs_avilable_space / units.Gi

                if free_space > vol_size:
                    return True
                break
        return False

    def _vrts_get_suitable_target(self, target_list, vol_size):
        """Get a suitable target for lun creation.

        Picking random target at first, if space is not available
        in first selected target then check each target one by one
        for suitable one.
        """

        target_count = len(target_list)

        incrmnt_pointer = 0
        target_index = randint(0, (target_count - 1))

        fs_list = self._vrts_get_fs_list()

        vrts_target_list = self._vrts_get_targets_store()

        store_name = self._vrts_get_assigned_store(
            target_list[target_index]['name'],
            vrts_target_list
        )

        if not self._vrts_is_space_available_in_store(
                vol_size, store_name, fs_list):
            while (incrmnt_pointer != target_count - 1):
                target_index = (target_index + 1) % target_count
                store_name = self._vrts_get_assigned_store(
                    target_list[target_index]['name'],
                    vrts_target_list
                )
                if self._vrts_is_space_available_in_store(
                        vol_size, store_name, fs_list):
                    return target_list[target_index]['name']
                incrmnt_pointer = incrmnt_pointer + 1
        else:
            return target_list[target_index]['name']

        return False

    def create_volume(self, volume):
        """Creates a Veritas Access Iscsi LUN."""
        create_dense = False
        if 'dense' in volume.metadata.keys():
            create_dense = strutils.bool_from_string(
                volume.metadata['dense'])

        lun_name = self._get_va_lun_name(volume.id)
        lun_size = '%sg' % volume.size
        path = self._lun_create_str
        provider = '%s:%s' % (self._va_ip, self._port)

        target_list = self._vrts_parse_xml_file(self.target_info_file)

        target_name = self._vrts_get_suitable_target(target_list, volume.size)

        if not target_name:
            message = (_('ACCESSIscsiDriver create volume failed %s '
                         'as no space is available') % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        data = {}
        data["lun_name"] = lun_name
        data["target_name"] = target_name
        data["size"] = lun_size
        if not self._sparse_lun_support or create_dense:
            data["option"] = "option=dense"

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver create volume failed %s')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def delete_volume(self, volume):
        """Deletes a Veritas Access Iscsi LUN."""
        lun_name = self._get_va_lun_name(volume.id)
        lun_list = self._get_vrts_lun_list()
        target_name = ""

        for lun in lun_list['output']['output']['luns']:
            if lun['lun_name'] == lun_name:
                target_name = lun['target_name']

        path = self._lun_destroy_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["target_name"] = target_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver delete volume failed %s')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def create_snapshot(self, snapshot):
        """Creates a snapshot of LUN."""
        lun_name = self._get_va_lun_name(snapshot.volume_id)
        snap_name = self._get_va_lun_name(snapshot.id)
        path = self._snapshot_create_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["snap_name"] = snap_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver create snapshot failed for %s')
                       % snapshot.volume_id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot of LUN."""
        lun_name = self._get_va_lun_name(snapshot.volume_id)
        snap_name = self._get_va_lun_name(snapshot.id)
        path = self._snapshot_destroy_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["snap_name"] = snap_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver delete snapshot failed for %s')
                       % snapshot.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the volume."""
        lun_name = self._get_va_lun_name(src_vref.id)
        cloned_lun_name = self._get_va_lun_name(volume.id)

        lun_found = False

        lun_list = self._get_vrts_lun_list()
        for lun in lun_list['output']['output']['luns']:
            if lun['lun_name'] == lun_name:
                store_name = lun['fs_name']
                lun_found = True
                break

        if not lun_found:
            message = (_('ACCESSIscsiDriver create cloned volume '
                         'failed %s as no source volume found') % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        fs_list = self._vrts_get_fs_list()

        if not self._vrts_is_space_available_in_store(volume.size, store_name,
                                                      fs_list):
            message = (_('ACCESSIscsiDriver create cloned volume '
                         'failed %s as no space is available') % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        path = self._lun_clone_create_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["clone_name"] = cloned_lun_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver create cloned '
                         'volume failed for %s')
                       % src_vref.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        if volume.size > src_vref.size:
            self._vrts_extend_lun(volume, volume.size)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from snapshot."""
        LOG.debug('ACCESSIscsiDriver create_volume_from_snapshot called')

        lun_name = self._get_va_lun_name(volume.id)
        snap_name = self._get_va_lun_name(snapshot.id)

        path = self._snapshot_list_str
        provider = '%s:%s' % (self._va_ip, self._port)
        data = {}
        data["snap_name"] = snap_name
        snap_info = self._access_api(self.session, provider, path,
                                     json.dumps(data), 'GET')

        target_name = ""
        if snap_info:
            for snap in snap_info['output']['output']['snapshots']:
                if snap['snapshot_name'] == snap_name:
                    target_name = snap['target_name']
                    break

        if target_name == "":
            message = (_('ACCESSIscsiDriver create volume from snapshot '
                         'failed for volume %s as failed to gather '
                         'snapshot details')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        vrts_target_list = self._vrts_get_targets_store()
        store_name = self._vrts_get_assigned_store(
            target_name, vrts_target_list)

        fs_list = self._vrts_get_fs_list()

        if not self._vrts_is_space_available_in_store(volume.size, store_name,
                                                      fs_list):
            message = (_('ACCESSIscsiDriver create volume from snapshot '
                         'failed %s as no space is available') % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        path = self._lun_create_from_snap_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["snap_name"] = snap_name

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')

        if not result:
            message = (_('ACCESSIscsiDriver create volume from snapshot '
                         'failed for volume %s')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        if volume.size > snapshot.volume_size:
            self._vrts_extend_lun(volume, volume.size)

    def _vrts_extend_lun(self, volume, size):
        """Extend vrts LUN to given size."""
        lun_name = self._get_va_lun_name(volume.id)
        target = {'target_name': ''}

        def _inner():
            lun_list = self._get_vrts_lun_list()
            for lun in lun_list['output']['output']['luns']:
                if lun['lun_name'] == lun_name:
                    target['target_name'] = lun['target_name']
                    raise loopingcall.LoopingCallDone()

        timer = loopingcall.FixedIntervalWithTimeoutLoopingCall(_inner)

        try:
            timer.start(interval=5, timeout=self.LUN_FOUND_INTERVAL).wait()
        except loopingcall.LoopingCallTimeOut:
            return False

        lun_size = '%sg' % size
        path = self._lun_extend_str
        provider = '%s:%s' % (self._va_ip, self._port)

        data = {}
        data["lun_name"] = lun_name
        data["target_name"] = target['target_name']
        data["size"] = lun_size

        result = self._access_api(self.session, provider, path,
                                  json.dumps(data), 'POST')
        return result

    def extend_volume(self, volume, size):
        """Extend the volume to new size"""
        lun_name = self._get_va_lun_name(volume.id)
        lun_found = False

        lun_list = self._get_vrts_lun_list()
        for lun in lun_list['output']['output']['luns']:
            if lun['lun_name'] == lun_name:
                store_name = lun['fs_name']
                lun_found = True
                break

        if not lun_found:
            message = (_('ACCESSIscsiDriver extend volume '
                         'failed %s as no volume found at backend')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        fs_list = self._vrts_get_fs_list()

        if not self._vrts_is_space_available_in_store(size, store_name,
                                                      fs_list):
            message = (_('ACCESSIscsiDriver extend volume '
                         'failed %s as no space is available') % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        result = self._vrts_extend_lun(volume, size)

        if not result:
            message = (_('ACCESSIscsiDriver extend '
                         'volume failed for %s')
                       % volume.id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def _get_api(self, provider, tail):
        api_root = 'https://%s/api/access' % (provider)
        if tail == self._fs_list_str:
            api_root = 'https://%s/api' % (provider)

        return api_root + tail

    def _access_api(self, session, provider, path, input_data, method):
        """Returns False if failure occurs."""
        kwargs = {'data': input_data}
        if not isinstance(input_data, dict):
            kwargs['headers'] = {'Content-Type': 'application/json'}
        full_url = self._get_api(provider, path)
        response = session.request(method, full_url, **kwargs)
        if response.status_code == 401:
            LOG.debug('Generating new session.')
            self.session = self._authenticate_access(self._va_ip,
                                                     self._user, self._pwd)
            response = self.session.request(method, full_url, **kwargs)

        if response.status_code != http_client.OK:
            LOG.error('Access API operation failed with HTTP error code %s.',
                      str(response.status_code))
            return False
        result = response.json()
        return result

    def _authenticate_access(self, address, username, password):
        session = requests.session()
        session.verify = self.verify
        session.auth = NoAuth()

        # Here 'address' will be only IPv4.
        response = session.post('https://%s:%s/api/rest/authenticate'
                                % (address, self._port),
                                data={'username': username,
                                      'password': password})
        if response.status_code != http_client.OK:
            LOG.error('Failed to authenticate to remote cluster at %s as %s.',
                      address, username)
            raise exception.NotAuthorized(_('Authentication failure.'))
        result = response.json()
        session.headers.update({'Authorization': 'Bearer {}'
                                .format(result['token'])})
        session.headers.update({'Content-Type': 'application/json'})

        return session

    def _get_va_backend_capacity(self):
        """Get VA backend total and free capacity."""
        target_list = self._vrts_parse_xml_file(self.target_info_file)
        fs_list = self._vrts_get_fs_list()
        vrts_target_list = self._vrts_get_targets_store()

        total_space = 0
        free_space = 0

        target_name = []
        target_store = []

        for target in target_list:
            target_name.append(target['name'])

        for target in vrts_target_list:
            if target['wwn'] in target_name:
                target_store.append(target['fs_list'][0])

        for store in target_store:
            for fs in fs_list:
                if fs['name'] == store:
                    total_space = total_space + fs['file_storage_capacity']
                    fs_free_space = (fs['file_storage_capacity'] -
                                     fs['file_storage_used'])

                    if fs_free_space > free_space:
                        free_space = fs_free_space

        total_capacity = int(total_space) / units.Gi
        free_capacity = int(free_space) / units.Gi

        return (total_capacity, free_capacity)

    def get_volume_stats(self, refresh=False):
        """Retrieve status info from share volume group."""

        total_capacity, free_capacity = self._get_va_backend_capacity()
        backend_name = self.configuration.safe_get('volume_backend_name')
        res_percentage = self.configuration.safe_get('reserved_percentage')
        self._stats["volume_backend_name"] = backend_name or 'VeritasISCSI'
        self._stats["vendor_name"] = 'Veritas'
        self._stats["reserved_percentage"] = res_percentage or 0
        self._stats["driver_version"] = self.VERSION
        self._stats["storage_protocol"] = self.DRIVER_VOLUME_TYPE
        self._stats['total_capacity_gb'] = total_capacity
        self._stats['free_capacity_gb'] = free_capacity
        self._stats['thin_provisioning_support'] = True

        return self._stats



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for Volume retype Code."""

from unittest import mock

from oslo_config import cfg

from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.policies import volume_actions as vol_action_policies
from cinder import quota
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import utils as tests_utils
from cinder.tests.unit import volume as base
from cinder.volume import volume_types


QUOTAS = quota.QUOTAS

CONF = cfg.CONF


class VolumeRetypeTestCase(base.BaseVolumeTestCase):
    """Verify multiattach retype restrictions."""

    def setUp(self):
        super(VolumeRetypeTestCase, self).setUp()
        self.patch('cinder.volume.volume_utils.clear_volume', autospec=True)
        self.expected_status = 'available'
        self.service_id = 1
        self.user_context = context.RequestContext(user_id=fake.USER_ID,
                                                   project_id=fake.PROJECT_ID)

        volume_types.create(self.context,
                            "fake_vol_type",
                            {},
                            description="fake_type")
        volume_types.create(self.context,
                            "fake_vol_type2",
                            {},
                            description="fake_type2")
        volume_types.create(self.context,
                            "multiattach-type",
                            {'multiattach': "<is> True"},
                            description="test-multiattach")
        volume_types.create(self.context,
                            "multiattach-type2",
                            {'multiattach': "<is> True"},
                            description="test-multiattach")
        self.default_vol_type = objects.VolumeType.get_by_name_or_id(
            self.context,
            'fake_vol_type')
        self.fake_vol_type2 = objects.VolumeType.get_by_name_or_id(
            self.context,
            'fake_vol_type2')
        self.multiattach_type = objects.VolumeType.get_by_name_or_id(
            self.context,
            'multiattach-type')
        self.multiattach_type2 = objects.VolumeType.get_by_name_or_id(
            self.context,
            'multiattach-type2')

    def fake_get_vtype(self, context, identifier):
        if identifier == "multiattach-type":
            return self.multiattach_type
        elif identifier == 'multiattach-type2':
            return self.multiattach_type2
        elif identifier == 'fake_vol_type2':
            return self.fake_vol_type2
        else:
            return self.default_vol_type

    @mock.patch('cinder.scheduler.rpcapi.SchedulerAPI.retype')
    @mock.patch('cinder.context.RequestContext.authorize')
    @mock.patch.object(volume_types, 'get_by_name_or_id')
    def test_retype_has_az(self, _mock_get_types, mock_authorize, mock_rpc):
        """Verify retype has az in request spec."""
        _mock_get_types.side_effect = self.fake_get_vtype

        vol = tests_utils.create_volume(
            self.context,
            volume_type_id=self.default_vol_type.id,
            status='available',
            availability_zone='nova')

        self.volume_api.retype(self.user_context,
                               vol,
                               'fake_vol_type2')

        mock_authorize.assert_called_once_with(
            vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY)

        fake_spec = {
            'volume_properties': mock.ANY,
            'volume_id': mock.ANY,
            'volume_type': mock.ANY,
            'migration_policy': mock.ANY,
            'quota_reservations': mock.ANY,
            'old_reservations': mock.ANY,
            'availability_zones': ['nova'],
        }

        mock_rpc.assert_called_once_with(
            self.user_context, mock.ANY,
            request_spec=fake_spec,
            filter_properties=mock.ANY
        )

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_non_multi_to_multi_retype(self, mock_authorize):
        """Test going from non-multiattach type to multiattach"""

        vol = tests_utils.create_volume(self.context,
                                        volume_type_id=
                                        self.default_vol_type.id)

        self.assertFalse(vol.multiattach)
        self.volume_api.retype(self.user_context,
                               vol,
                               'multiattach-type')
        vol.refresh()
        self.assertTrue(vol.multiattach)

        mock_authorize.assert_has_calls(
            [mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY)
             ])

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_multi_to_non_multi_retype(self, mock_authorize):
        """Test going from multiattach to a non-multiattach type"""

        vol = tests_utils.create_volume(self.context,
                                        multiattach=True,
                                        volume_type_id=
                                        self.multiattach_type.id)

        self.assertTrue(vol.multiattach)
        self.volume_api.retype(self.user_context,
                               vol,
                               'fake_vol_type')
        vol.refresh()
        self.assertFalse(vol.multiattach)

        mock_authorize.assert_has_calls(
            [mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY)
             ])

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_in_use_volume_retype(self, mock_authorize):
        """Test trying to retype an in-use volume"""

        vol = tests_utils.create_volume(self.context,
                                        volume_type_id=
                                        self.multiattach_type.id)
        vol.update({'status': 'in-use'})
        vol.save()
        self.assertRaises(exception.InvalidInput,
                          self.volume_api.retype,
                          self.context,
                          vol,
                          'multiattach-type')
        mock_authorize.assert_has_calls(
            [mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY),
             ])

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_multiattach_to_multiattach_retype(self, mock_authorize):
        """Test going from multiattach to multiattach"""

        vol = tests_utils.create_volume(self.context,
                                        multiattach=True,
                                        volume_type_id=
                                        self.multiattach_type.id)

        self.assertTrue(vol.multiattach)
        self.volume_api.retype(self.user_context,
                               vol,
                               'multiattach-type2')
        vol.refresh()
        self.assertTrue(vol.multiattach)

        mock_authorize.assert_has_calls(
            [mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY)
             ])

    def test_retype_driver_not_initialized(self):
        volume = tests_utils.create_volume(
            self.context,
            host=CONF.host,
            status='available',
            volume_type_id=self.default_vol_type.id)

        host_obj = {'host': CONF.host, 'capabilities': {}}

        self.volume.driver._initialized = False
        self.assertRaises(exception.DriverNotInitialized,
                          self.volume.retype,
                          self.context, volume,
                          self.multiattach_type.id, host_obj,
                          migration_policy='on-demand')

        volume.refresh()
        self.assertEqual('available', volume.status)

    def test_retype_with_volume_type_resize_limits(self):

        def _create_min_max_size_dict(min_size, max_size):
            return {volume_types.MIN_SIZE_KEY: min_size,
                    volume_types.MAX_SIZE_KEY: max_size}

        def _setup_volume_types():
            spec_dict = _create_min_max_size_dict(2, 4)
            sized_vol_type_dict = {'name': 'limit_type',
                                   'extra_specs': spec_dict}
            db.volume_type_create(self.context, sized_vol_type_dict)
            self.sized_vol_type = db.volume_type_get_by_name(
                self.context, sized_vol_type_dict['name'])

            unsized_vol_type_dict = {'name': 'unsized_type', 'extra_specs': {}}
            db.volume_type_create(context.get_admin_context(),
                                  unsized_vol_type_dict)
            self.unsized_vol_type = db.volume_type_get_by_name(
                self.context, unsized_vol_type_dict['name'])

        _setup_volume_types()
        volume_1 = tests_utils.create_volume(
            self.context,
            host=CONF.host,
            status='available',
            volume_type_id=self.default_vol_type.id,
            size=1)
        volume_3 = tests_utils.create_volume(
            self.context,
            host=CONF.host,
            status='available',
            volume_type_id=self.default_vol_type.id,
            size=3)
        volume_9 = tests_utils.create_volume(
            self.context,
            host=CONF.host,
            status='available',
            volume_type_id=self.default_vol_type.id,
            size=9)

        self.assertRaises(exception.InvalidInput,
                          self.volume_api.retype,
                          self.context, volume_1,
                          'limit_type',
                          migration_policy='on-demand')
        self.assertRaises(exception.InvalidInput,
                          self.volume_api.retype,
                          self.context, volume_9,
                          'limit_type',
                          migration_policy='on-demand')
        self.volume_api.retype(self.context, volume_3,
                               'limit_type', migration_policy='on-demand')



# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volumes api."""


from oslo_config import cfg
from oslo_log import log as logging
from oslo_log import versionutils
from oslo_utils import uuidutils
from six.moves import http_client
import webob
from webob import exc

from cinder.api import api_utils
from cinder.api import common
from cinder.api.contrib import scheduler_hints
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import volumes
from cinder.api.v2.views import volumes as volume_views
from cinder.api import validation
from cinder import exception
from cinder import group as group_api
from cinder.i18n import _
from cinder.image import glance
from cinder import objects
from cinder import utils
from cinder import volume as cinder_volume
from cinder.volume import volume_utils

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class VolumeController(wsgi.Controller):
    """The Volumes API controller for the OpenStack API."""

    _view_builder_class = volume_views.ViewBuilder

    def __init__(self, ext_mgr):
        self.volume_api = cinder_volume.API()
        self.group_api = group_api.API()
        self.ext_mgr = ext_mgr
        super(VolumeController, self).__init__()

    def show(self, req, id):
        """Return data about the given volume."""
        context = req.environ['cinder.context']

        # Not found exception will be handled at the wsgi level
        vol = self.volume_api.get(context, id, viewable_admin_meta=True)
        req.cache_db_volume(vol)

        api_utils.add_visible_admin_metadata(vol)

        return self._view_builder.detail(req, vol)

    def delete(self, req, id):
        """Delete a volume."""
        context = req.environ['cinder.context']

        cascade = utils.get_bool_param('cascade', req.params)

        LOG.info("Delete volume with id: %s", id)

        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)
        self.volume_api.delete(context, volume, cascade=cascade)
        return webob.Response(status_int=http_client.ACCEPTED)

    def index(self, req):
        """Returns a summary list of volumes."""
        return self._get_volumes(req, is_detail=False)

    def detail(self, req):
        """Returns a detailed list of volumes."""
        return self._get_volumes(req, is_detail=True)

    def _get_volumes(self, req, is_detail):
        """Returns a list of volumes, transformed through view builder."""

        context = req.environ['cinder.context']

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        # NOTE(wanghao): Always removing glance_metadata since we support it
        # only in API version >= VOLUME_LIST_GLANCE_METADATA.
        filters.pop('glance_metadata', None)
        api_utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_volume_filter_options())

        # NOTE(thingee): v2 API allows name instead of display_name
        if 'name' in sort_keys:
            sort_keys[sort_keys.index('name')] = 'display_name'

        if 'name' in filters:
            filters['display_name'] = filters.pop('name')

        self.volume_api.check_volume_filters(filters)
        volumes = self.volume_api.get_all(context, marker, limit,
                                          sort_keys=sort_keys,
                                          sort_dirs=sort_dirs,
                                          filters=filters,
                                          viewable_admin_meta=True,
                                          offset=offset)

        for volume in volumes:
            api_utils.add_visible_admin_metadata(volume)

        req.cache_db_volumes(volumes.objects)

        if is_detail:
            volumes = self._view_builder.detail_list(req, volumes)
        else:
            volumes = self._view_builder.summary_list(req, volumes)
        return volumes

    def _image_uuid_from_ref(self, image_ref, context):
        # If the image ref was generated by nova api, strip image_ref
        # down to an id.
        image_uuid = None
        try:
            image_uuid = image_ref.split('/').pop()
        except AttributeError:
            msg = _("Invalid imageRef provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        image_service = glance.get_default_image_service()

        # First see if this is an actual image ID
        if uuidutils.is_uuid_like(image_uuid):
            try:
                image = image_service.show(context, image_uuid)
                if 'id' in image:
                    return image['id']
            except Exception:
                # Pass and see if there is a matching image name
                pass

        # Could not find by ID, check if it is an image name
        try:
            params = {'filters': {'name': image_ref}}
            images = list(image_service.detail(context, **params))
            if len(images) > 1:
                msg = _("Multiple matches found for '%s', use an ID to be more"
                        " specific.") % image_ref
                raise exc.HTTPConflict(explanation=msg)
            for img in images:
                return img['id']
        except exc.HTTPConflict:
            raise
        except Exception:
            # Pass the other exception and let default not found error
            # handling take care of it
            pass

        msg = _("Invalid image identifier or unable to "
                "access requested image.")
        raise exc.HTTPBadRequest(explanation=msg)

    @wsgi.response(http_client.ACCEPTED)
    @validation.schema(volumes.create, mv.V2_BASE_VERSION)
    def create(self, req, body):
        """Creates a new volume."""

        LOG.debug('Create volume request body: %s', body)
        context = req.environ['cinder.context']

        # NOTE (pooja_jadhav) To fix bug 1774155, scheduler hints is not
        # loaded as a standard extension. If user passes
        # OS-SCH-HNT:scheduler_hints in the request body, then it will be
        # validated in the create method and this method will add
        # scheduler_hints in body['volume'].
        body = scheduler_hints.create(req, body)
        volume = body['volume']

        kwargs = {}
        self.validate_name_and_description(volume, check_length=False)

        # NOTE(thingee): v2 API allows name instead of display_name
        if 'name' in volume:
            volume['display_name'] = volume.pop('name')

        # NOTE(thingee): v2 API allows description instead of
        #                display_description
        if 'description' in volume:
            volume['display_description'] = volume.pop('description')

        if 'image_id' in volume:
            volume['imageRef'] = volume.pop('image_id')

        req_volume_type = volume.get('volume_type', None)
        if req_volume_type:
            # Not found exception will be handled at the wsgi level
            kwargs['volume_type'] = (
                objects.VolumeType.get_by_name_or_id(context, req_volume_type))

        kwargs['metadata'] = volume.get('metadata', None)

        snapshot_id = volume.get('snapshot_id')
        if snapshot_id is not None:
            # Not found exception will be handled at the wsgi level
            kwargs['snapshot'] = self.volume_api.get_snapshot(context,
                                                              snapshot_id)
        else:
            kwargs['snapshot'] = None

        source_volid = volume.get('source_volid')
        if source_volid is not None:
            # Not found exception will be handled at the wsgi level
            kwargs['source_volume'] = \
                self.volume_api.get_volume(context,
                                           source_volid)
        else:
            kwargs['source_volume'] = None

        kwargs['group'] = None
        kwargs['consistencygroup'] = None
        consistencygroup_id = volume.get('consistencygroup_id')
        if consistencygroup_id is not None:
            # Not found exception will be handled at the wsgi level
            kwargs['group'] = self.group_api.get(context, consistencygroup_id)

        size = volume.get('size', None)
        if size is None and kwargs['snapshot'] is not None:
            size = kwargs['snapshot']['volume_size']
        elif size is None and kwargs['source_volume'] is not None:
            size = kwargs['source_volume']['size']

        LOG.info("Create volume of %s GB", size)

        image_ref = volume.get('imageRef')
        if image_ref is not None:
            image_uuid = self._image_uuid_from_ref(image_ref, context)
            kwargs['image_id'] = image_uuid

        kwargs['availability_zone'] = volume.get('availability_zone', None)
        kwargs['scheduler_hints'] = volume.get('scheduler_hints', None)
        kwargs['multiattach'] = utils.get_bool_param('multiattach', volume)

        if kwargs.get('multiattach', False):
            msg = ("The option 'multiattach' "
                   "is deprecated and will be removed in a future "
                   "release.  The default behavior going forward will "
                   "be to specify multiattach enabled volume types.")
            versionutils.report_deprecated_feature(LOG, msg)

        try:
            new_volume = self.volume_api.create(
                context, size, volume.get('display_name'),
                volume.get('display_description'), **kwargs)
        except exception.VolumeTypeDefaultMisconfiguredError as err:
            raise webob.exc.HTTPInternalServerError(explanation=err.msg)

        retval = self._view_builder.detail(req, new_volume)

        return retval

    def _get_volume_filter_options(self):
        """Return volume search options allowed by non-admin."""
        return common.get_enabled_resource_filters('volume').get('volume', [])

    @validation.schema(volumes.update, mv.V2_BASE_VERSION,
                       mv.get_prior_version(mv.SUPPORT_VOLUME_SCHEMA_CHANGES))
    @validation.schema(volumes.update_volume_v353,
                       mv.SUPPORT_VOLUME_SCHEMA_CHANGES)
    def update(self, req, id, body):
        """Update a volume."""
        context = req.environ['cinder.context']
        update_dict = body['volume']

        self.validate_name_and_description(update_dict, check_length=False)

        # NOTE(thingee): v2 API allows name instead of display_name
        if 'name' in update_dict:
            update_dict['display_name'] = update_dict.pop('name')

        # NOTE(thingee): v2 API allows description instead of
        #                display_description
        if 'description' in update_dict:
            update_dict['display_description'] = update_dict.pop('description')

        # Not found and Invalid exceptions will be handled at the wsgi level
        try:
            volume = self.volume_api.get(context, id, viewable_admin_meta=True)
            volume_utils.notify_about_volume_usage(context, volume,
                                                   'update.start')
            self.volume_api.update(context, volume, update_dict)
        except exception.InvalidVolumeMetadataSize as error:
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=error.msg)

        volume.update(update_dict)

        api_utils.add_visible_admin_metadata(volume)

        volume_utils.notify_about_volume_usage(context, volume,
                                               'update.end')

        return self._view_builder.detail(req, volume)


def create_resource(ext_mgr):
    return wsgi.Resource(VolumeController(ext_mgr))



# Copyright (C) 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy


group_specs_with_no_spaces_key_and_value_null = {
    'type': 'object',
    'patternProperties': {
        '^[a-zA-Z0-9-_:.]{1,255}$': {
            'type': ['string', 'null'], 'maxLength': 255
        }
    },
    'additionalProperties': False
}

create = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'group_specs': group_specs_with_no_spaces_key_and_value_null,
    },
    'required': ['group_specs'],
    'additionalProperties': False,
}

update = copy.deepcopy(group_specs_with_no_spaces_key_and_value_null)
update.update({
    'minProperties': 1,
    'maxProperties': 1
})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.tests.functional.api_sample_tests import fakes
from cinder.tests.functional import api_samples_test_base as base


class SnapshotManageExtensionsSampleJsonTest(base.ApiSampleTestBase):
    sample_dir = "snapshot_manage_extensions"

    def setUp(self):
        super(SnapshotManageExtensionsSampleJsonTest, self).setUp()
        self.stub_out("cinder.api.contrib.snapshot_manage."
                      "SnapshotManageController.create",
                      fakes.stub_manage_existing_snapshot)

    def test_snapshot_manage_create(self):

        response = self._do_post('os-snapshot-manage',
                                 'snapshot-manage-request',
                                 subs=self.subs)
        self._verify_response('snapshot-manage-response',
                              {}, response, 202)



# Copyright 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from cinder.objects import fields
from cinder.tests.functional import functional_helpers
from cinder.volume import configuration


class GroupSnapshotsTest(functional_helpers._FunctionalTestBase):
    _vol_type_name = 'functional_test_type'
    _grp_type_name = 'functional_grp_test_type'
    osapi_version_major = '3'
    osapi_version_minor = '19'

    def setUp(self):
        super(GroupSnapshotsTest, self).setUp()
        self.volume_type = self.api.create_type(self._vol_type_name)
        self.group_type = self.api.create_group_type(self._grp_type_name)

    def _get_flags(self):
        f = super(GroupSnapshotsTest, self)._get_flags()
        f['volume_driver'] = (
            {'v': 'cinder.tests.fake_driver.FakeLoggingVolumeDriver',
             'g': configuration.SHARED_CONF_GROUP})
        f['default_volume_type'] = {'v': self._vol_type_name}
        f['default_group_type'] = {'v': self._grp_type_name}
        return f

    def test_get_group_snapshots_summary(self):
        """Simple check that listing group snapshots works."""
        grp_snaps = self.api.get_group_snapshots(False)
        self.assertIsNotNone(grp_snaps)

    def test_get_group_snapshots(self):
        """Simple check that listing group snapshots works."""
        grp_snaps = self.api.get_group_snapshots()
        self.assertIsNotNone(grp_snaps)

    def test_create_and_delete_group_snapshot(self):
        """Creates and deletes a group snapshot."""

        # Create group
        created_group = self.api.post_group(
            {'group': {'group_type': self.group_type['id'],
                       'volume_types': [self.volume_type['id']]}})
        self.assertTrue(uuidutils.is_uuid_like(created_group['id']))
        created_group_id = created_group['id']

        # Check it's there
        found_group = self._poll_group_while(created_group_id,
                                             ['creating'])
        self.assertEqual(created_group_id, found_group['id'])
        self.assertEqual(self.group_type['id'], found_group['group_type'])
        self.assertEqual('available', found_group['status'])

        # Create volume
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'group_id': created_group_id,
                        'volume_type': self.volume_type['id']}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(self._vol_type_name, found_volume['volume_type'])
        self.assertEqual(created_group_id, found_volume['group_id'])

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])

        # It should be available...
        self.assertEqual('available', found_volume['status'])

        # Create group snapshot
        created_group_snapshot = self.api.post_group_snapshot(
            {'group_snapshot': {'group_id': created_group_id}})
        self.assertTrue(uuidutils.is_uuid_like(created_group_snapshot['id']))
        created_group_snapshot_id = created_group_snapshot['id']

        # Check it's there
        found_group_snapshot = self._poll_group_snapshot_while(
            created_group_snapshot_id, [fields.GroupSnapshotStatus.CREATING])
        self.assertEqual(created_group_snapshot_id, found_group_snapshot['id'])
        self.assertEqual(created_group_id,
                         found_group_snapshot['group_id'])
        self.assertEqual(fields.GroupSnapshotStatus.AVAILABLE,
                         found_group_snapshot['status'])

        # Delete the group snapshot
        self.api.delete_group_snapshot(created_group_snapshot_id)

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_group_snapshot = self._poll_group_snapshot_while(
            created_group_snapshot_id, [fields.GroupSnapshotStatus.DELETING])

        # Delete the original group
        self.api.delete_group(created_group_id,
                              {'delete': {'delete-volumes': True}})

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        found_group = self._poll_group_while(created_group_id, ['deleting'])

        # Should be gone
        self.assertIsNone(found_group_snapshot)
        self.assertIsNone(found_volume)
        self.assertIsNone(found_group)

    def test_create_group_from_group_snapshot(self):
        """Creates a group from a group snapshot."""

        # Create group
        created_group = self.api.post_group(
            {'group': {'group_type': self.group_type['id'],
                       'volume_types': [self.volume_type['id']]}})
        self.assertTrue(uuidutils.is_uuid_like(created_group['id']))
        created_group_id = created_group['id']

        # Check it's there
        found_group = self._poll_group_while(created_group_id,
                                             ['creating'])
        self.assertEqual(created_group_id, found_group['id'])
        self.assertEqual(self.group_type['id'], found_group['group_type'])
        self.assertEqual('available', found_group['status'])

        # Create volume
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'group_id': created_group_id,
                        'volume_type': self.volume_type['id']}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(self._vol_type_name, found_volume['volume_type'])
        self.assertEqual(created_group_id, found_volume['group_id'])

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])

        # It should be available...
        self.assertEqual('available', found_volume['status'])

        # Create group snapshot
        created_group_snapshot = self.api.post_group_snapshot(
            {'group_snapshot': {'group_id': created_group_id}})
        self.assertTrue(uuidutils.is_uuid_like(created_group_snapshot['id']))
        created_group_snapshot_id = created_group_snapshot['id']

        # Check it's there
        found_group_snapshot = self._poll_group_snapshot_while(
            created_group_snapshot_id, ['creating'])
        self.assertEqual(created_group_snapshot_id, found_group_snapshot['id'])
        self.assertEqual(created_group_id,
                         found_group_snapshot['group_id'])
        self.assertEqual('available', found_group_snapshot['status'])

        # Create group from group snapshot
        created_group_from_snap = self.api.post_group_from_src(
            {'create-from-src': {
                'group_snapshot_id': created_group_snapshot_id}})
        self.assertTrue(uuidutils.is_uuid_like(created_group_from_snap['id']))
        created_group_from_snap_id = created_group_from_snap['id']

        # Check it's there
        found_volumes = self.api.get_volumes()
        self._poll_volume_while(found_volumes[0], ['creating'])
        self._poll_volume_while(found_volumes[1], ['creating'])
        found_group_from_snap = self._poll_group_while(
            created_group_from_snap_id, ['creating'])
        self.assertEqual(created_group_from_snap_id,
                         found_group_from_snap['id'])
        self.assertEqual(created_group_snapshot_id,
                         found_group_from_snap['group_snapshot_id'])
        self.assertEqual(self.group_type['id'],
                         found_group_from_snap['group_type'])
        self.assertEqual('available', found_group_from_snap['status'])

        # Delete the group from snap
        self.api.delete_group(created_group_from_snap_id,
                              {'delete': {'delete-volumes': True}})

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_group_from_snap = self._poll_group_while(
            created_group_from_snap_id, ['deleting'])

        # Delete the group snapshot
        self.api.delete_group_snapshot(created_group_snapshot_id)

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_group_snapshot = self._poll_group_snapshot_while(
            created_group_snapshot_id, [fields.GroupSnapshotStatus.DELETING])

        # Delete the original group
        self.api.delete_group(created_group_id,
                              {'delete': {'delete-volumes': True}})

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        found_group = self._poll_group_while(created_group_id, ['deleting'])

        # Should be gone
        self.assertIsNone(found_group_from_snap)
        self.assertIsNone(found_group_snapshot)
        self.assertIsNone(found_volume)
        self.assertIsNone(found_group)

    def test_create_group_from_source_group(self):
        """Creates a group from a source group."""

        # Create group
        created_group = self.api.post_group(
            {'group': {'group_type': self.group_type['id'],
                       'volume_types': [self.volume_type['id']]}})
        self.assertTrue(uuidutils.is_uuid_like(created_group['id']))
        created_group_id = created_group['id']

        # Check it's there
        found_group = self._poll_group_while(created_group_id,
                                             ['creating'])
        self.assertEqual(created_group_id, found_group['id'])
        self.assertEqual(self.group_type['id'], found_group['group_type'])
        self.assertEqual('available', found_group['status'])

        # Create volume
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'group_id': created_group_id,
                        'volume_type': self.volume_type['id']}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(self._vol_type_name, found_volume['volume_type'])
        self.assertEqual(created_group_id, found_volume['group_id'])

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])

        # It should be available...
        self.assertEqual('available', found_volume['status'])

        # Test create group from source group
        created_group_from_group = self.api.post_group_from_src(
            {'create-from-src': {
                'source_group_id': created_group_id}})
        self.assertTrue(uuidutils.is_uuid_like(created_group_from_group['id']))
        created_group_from_group_id = created_group_from_group['id']

        # Check it's there
        found_volumes = self.api.get_volumes()
        self._poll_volume_while(found_volumes[0], ['creating'])
        self._poll_volume_while(found_volumes[1], ['creating'])
        found_group_from_group = self._poll_group_while(
            created_group_from_group_id, ['creating'])
        self.assertEqual(created_group_from_group_id,
                         found_group_from_group['id'])
        self.assertEqual(created_group_id,
                         found_group_from_group['source_group_id'])
        self.assertEqual(self.group_type['id'],
                         found_group_from_group['group_type'])
        self.assertEqual('available', found_group_from_group['status'])

        # Delete the group from group
        self.api.delete_group(created_group_from_group_id,
                              {'delete': {'delete-volumes': True}})

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_group_from_group = self._poll_group_while(
            created_group_from_group_id, ['deleting'])

        # Delete the original group
        self.api.delete_group(created_group_id,
                              {'delete': {'delete-volumes': True}})

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        found_group = self._poll_group_while(created_group_id, ['deleting'])

        # Should be gone
        self.assertIsNone(found_group_from_group)
        self.assertIsNone(found_volume)
        self.assertIsNone(found_group)

    def test_reset_group_snapshot(self):
        # Create group
        group1 = self.api.post_group(
            {'group': {'group_type': self.group_type['id'],
                       'volume_types': [self.volume_type['id']]}})
        self.assertTrue(uuidutils.is_uuid_like(group1['id']))
        group_id = group1['id']
        self._poll_group_while(group_id, ['creating'])

        # Create volume
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'group_id': group_id,
                        'volume_type': self.volume_type['id']}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']
        self._poll_volume_while(created_volume_id, ['creating'])

        # Create group snapshot
        group_snapshot1 = self.api.post_group_snapshot(
            {'group_snapshot': {'group_id': group_id}})
        self.assertTrue(uuidutils.is_uuid_like(group_snapshot1['id']))
        group_snapshot_id = group_snapshot1['id']

        self._poll_group_snapshot_while(group_snapshot_id,
                                        fields.GroupSnapshotStatus.CREATING)

        group_snapshot1 = self.api.get_group_snapshot(group_snapshot_id)
        self.assertEqual(fields.GroupSnapshotStatus.AVAILABLE,
                         group_snapshot1['status'])

        # reset group snapshot status
        self.api.reset_group_snapshot(group_snapshot_id, {"reset_status": {
            "status": fields.GroupSnapshotStatus.ERROR}})

        group_snapshot1 = self.api.get_group_snapshot(group_snapshot_id)
        self.assertEqual(fields.GroupSnapshotStatus.ERROR,
                         group_snapshot1['status'])

        # Delete group, volume and group snapshot
        self.api.delete_group_snapshot(group_snapshot_id)
        found_group_snapshot = self._poll_group_snapshot_while(
            group_snapshot_id, [fields.GroupSnapshotStatus.DELETING])
        self.api.delete_group(group_id,
                              {'delete': {'delete-volumes': True}})

        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        found_group = self._poll_group_while(group_id, ['deleting'])

        # Created resources should be gone
        self.assertIsNone(found_group_snapshot)
        self.assertIsNone(found_volume)
        self.assertIsNone(found_group)
